from os import path


def resolve_file_path(file_path: str, relative_to_file: str = None, relative_to_dir: str = None) -> str:
    """
    Resolve a relative file path (or return absolute ones as-is).
    :param file_path:         the relative path
    :param relative_to_file:  a file this path is relative to (note: this is a file, so we
                              will use the file's dirname, not the file itself)
    :param relative_to_dir:   a directory this path is relative to
    :return: a full path name
    """
    if path.isabs(file_path):
        return file_path
    if relative_to_file is not None:
        return path.join(path.dirname(relative_to_file), file_path)
    if relative_to_dir is None:
        raise ValueError('resolve_file_path requires either relative_to_file or relative_to_dir arg!')
    return path.join(relative_to_dir, file_path)



from semantic_version import Version

from .. import semver as sut


def test_normalize():
    assert '1.1.0' == sut.normalize('v1.1.0')
    assert '1.1.0' == sut.normalize('1.1.0')


def test_parse():
    assert [1, 1, 0, (), ()] == list((sut.parse('v1.1.0')))
    assert [1, 1, 0, ('2',), ('build3',)] == list((sut.parse('v1.1.0-2+build3')))
    assert [2, 0, 0, (), ()] == list(sut.parse('2.0'))



from .. import helm as sut


def test_lookup_value():
    assert 'foo' == sut.lookup_value('a.b', {'a': {'b': 'foo'}})
    assert 'foo' == sut.lookup_value(['a', 'b'], {'a': {'b': 'foo'}})
    assert sut.lookup_value(['b.a'], {'a': {'b': 'foo'}}) is None


def test_key_is_in_values():
    assert sut.key_is_in_values(['image', 'tag'], {'image': {'tag': 'foo'}})
    assert sut.key_is_in_values('image.tag', {'image': {'tag': 'foo'}})
    assert not sut.key_is_in_values(['image', 'tag'], {'imageTag': 'foo'})
    assert not sut.key_is_in_values(['image.tag'], {'imageTag': 'foo'})



from typing import List

from .gen_py.ttypes import Environment, Cluster


def generate_environment_init_command(cluster: Cluster, env: Environment, dry_run: bool = False) -> List[List[str]]:
    """
    :param cluster:  cluster object for the environment
    :param env:      which environment we should generate init command[s] for
    :param dry_run:  if True, do not generate mutating commands
    :return: a list of commands, each command is an argv list
    """
    commands = []
    context_name = 'env:' + env.name
    if cluster.provider.gke:
        commands.append(['gcloud', 'container', 'clusters', 'get-credentials',
                         '--project', cluster.provider.gke.project,
                         '--zone', cluster.provider.gke.zone,
                         cluster.provider.gke.clusterName])
        cluster_name = 'gke_{project}_{zone}_{cluster}'.format(project=cluster.provider.gke.project,
                                                               zone=cluster.provider.gke.zone,
                                                               cluster=cluster.name)
        commands.append(['kubectl', 'config', 'set-context', context_name,
                         '--cluster', cluster_name,
                         '--user', cluster_name,
                         '--namespace', env.kubeNamespace])
    elif cluster.provider.minikube:
        cluster_name = 'minikube'
        commands.append(['kubectl', 'config', 'set-context', context_name,
                         '--cluster', cluster_name,
                         '--user', cluster_name,
                         '--namespace', env.kubeNamespace])
    return commands



#
# Autogenerated by Thrift Compiler (0.11.0)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:dynamic
#

from thrift.Thrift import TType, TMessageType, TFrozenDict, TException, TApplicationException
from thrift.protocol.TProtocol import TProtocolException
from thrift.TRecursive import fix_spec

import sys

from thrift.transport import TTransport
from thrift.protocol.TBase import TBase, TFrozenBase, TExceptionBase, TTransport
all_structs = []


class NameFromRef(TBase):
    """
    Attributes:
     - clusterParam
    """


    def __init__(self, clusterParam=None,):
        self.clusterParam = clusterParam

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class GceAddressValueRef(TBase):
    """
    Attributes:
     - name
     - nameFrom
     - isGlobal
    """


    def __init__(self, name=None, nameFrom=None, isGlobal=None,):
        self.name = name
        self.nameFrom = nameFrom
        self.isGlobal = isGlobal

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class GceValueRef(TBase):
    """
    Attributes:
     - address
    """


    def __init__(self, address=None,):
        self.address = address

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class ChartValueRef(TBase):
    """
    Attributes:
     - gceResource
    """


    def __init__(self, gceResource=None,):
        self.gceResource = gceResource

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class ChartValue(TBase):
    """
    Attributes:
     - key
     - value
     - valueFrom
    """


    def __init__(self, key=None, value=None, valueFrom=None,):
        self.key = key
        self.value = value
        self.valueFrom = valueFrom

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class GithubTrigger(TBase):
    """
    Attributes:
     - repository
     - branch
     - tagMatching
    """


    def __init__(self, repository=None, branch=None, tagMatching=None,):
        self.repository = repository
        self.branch = branch
        self.tagMatching = tagMatching

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class ImageTrigger(TBase):
    """
    Attributes:
     - tagValue
     - repoValue
     - repoPrefixValue
     - track
    """


    def __init__(self, tagValue="image.tag", repoValue="image.repository", repoPrefixValue="image.prefix", track=None,):
        self.tagValue = tagValue
        self.repoValue = repoValue
        self.repoPrefixValue = repoPrefixValue
        self.track = track

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class HelmTrigger(TBase):
    """
    Attributes:
     - track
    """


    def __init__(self, track=None,):
        self.track = track

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class DeploymentTrigger(TBase):
    """
    Attributes:
     - github
     - image
     - helm
    """


    def __init__(self, github=None, image=None, helm=None,):
        self.github = github
        self.image = image
        self.helm = helm

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Chart(TBase):
    """
    Attributes:
     - reference
     - dir
     - version
    """


    def __init__(self, reference=None, dir=None, version=None,):
        self.reference = reference
        self.dir = dir
        self.version = version

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Release(TBase):
    """
    Attributes:
     - name
     - chart
     - valuesFile
     - values
     - trigger
     - triggers
     - skipDefaultValues
     - resourceFiles
    """


    def __init__(self, name=None, chart=None, valuesFile=None, values=None, trigger=None, triggers=None, skipDefaultValues=None, resourceFiles=None,):
        self.name = name
        self.chart = chart
        self.valuesFile = valuesFile
        self.values = values
        self.trigger = trigger
        self.triggers = triggers
        self.skipDefaultValues = skipDefaultValues
        self.resourceFiles = resourceFiles

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class KubernetesResourceRef(TBase):
    """
    Attributes:
     - kind
     - name
    """


    def __init__(self, kind=None, name=None,):
        self.kind = kind
        self.name = name

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Releases(TBase):
    """
    Attributes:
     - resourceFiles
     - releases
     - resourceDependencies
    """


    def __init__(self, resourceFiles=None, releases=None, resourceDependencies=None,):
        self.resourceFiles = resourceFiles
        self.releases = releases
        self.resourceDependencies = resourceDependencies

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class GkeProvider(TBase):
    """
    Attributes:
     - project
     - clusterName
     - zone
    """


    def __init__(self, project=None, clusterName=None, zone=None,):
        self.project = project
        self.clusterName = clusterName
        self.zone = zone

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class MinikubeProvider(TBase):


    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Provider(TBase):
    """
    Attributes:
     - gke
     - minikube
    """


    def __init__(self, gke=None, minikube=None,):
        self.gke = gke
        self.minikube = minikube

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class ClusterParameter(TBase):
    """
    Attributes:
     - name
     - value
    """


    def __init__(self, name=None, value=None,):
        self.name = name
        self.value = value

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Cluster(TBase):
    """
    Attributes:
     - name
     - provider
     - parameters
    """


    def __init__(self, name=None, provider=None, parameters=None,):
        self.name = name
        self.provider = provider
        self.parameters = parameters

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class Environment(TBase):
    """
    Attributes:
     - name
     - clusterName
     - kubeNamespace
     - releasesFiles: a list of `releases.yaml` files
     - defaultValuesFile: default helm values file for the environment
     - defaultValues
    """


    def __init__(self, name=None, clusterName=None, kubeNamespace=None, releasesFiles=None, defaultValuesFile=None, defaultValues=None,):
        self.name = name
        self.clusterName = clusterName
        self.kubeNamespace = kubeNamespace
        self.releasesFiles = releasesFiles
        self.defaultValuesFile = defaultValuesFile
        self.defaultValues = defaultValues

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class HelmRepo(TBase):
    """
    Attributes:
     - name
     - url
     - caFile
     - certFile
     - keyFile
    """


    def __init__(self, name=None, url=None, caFile=None, certFile=None, keyFile=None,):
        self.name = name
        self.url = url
        self.caFile = caFile
        self.certFile = certFile
        self.keyFile = keyFile

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)


class KubecdConfig(TBase):
    """
    Attributes:
     - clusters
     - environments
     - helmRepos
    """


    def __init__(self, clusters=None, environments=None, helmRepos=None,):
        self.clusters = clusters
        self.environments = environments
        self.helmRepos = helmRepos

    def __repr__(self):
        L = ['%s=%r' % (key, value)
             for key, value in self.__dict__.items()]
        return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not (self == other)
all_structs.append(NameFromRef)
NameFromRef.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'clusterParam', 'UTF8', None, ),  # 1
)
all_structs.append(GceAddressValueRef)
GceAddressValueRef.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'name', 'UTF8', None, ),  # 1
    (2, TType.STRUCT, 'nameFrom', [NameFromRef, None], None, ),  # 2
    (3, TType.BOOL, 'isGlobal', None, None, ),  # 3
)
all_structs.append(GceValueRef)
GceValueRef.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'address', [GceAddressValueRef, None], None, ),  # 1
)
all_structs.append(ChartValueRef)
ChartValueRef.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'gceResource', [GceValueRef, None], None, ),  # 1
)
all_structs.append(ChartValue)
ChartValue.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'key', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'value', 'UTF8', None, ),  # 2
    (3, TType.STRUCT, 'valueFrom', [ChartValueRef, None], None, ),  # 3
)
all_structs.append(GithubTrigger)
GithubTrigger.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'repository', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'branch', 'UTF8', None, ),  # 2
    (3, TType.STRING, 'tagMatching', 'UTF8', None, ),  # 3
)
all_structs.append(ImageTrigger)
ImageTrigger.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'tagValue', 'UTF8', "image.tag", ),  # 1
    (2, TType.STRING, 'repoValue', 'UTF8', "image.repository", ),  # 2
    (3, TType.STRING, 'repoPrefixValue', 'UTF8', "image.prefix", ),  # 3
    (4, TType.STRING, 'track', 'UTF8', None, ),  # 4
)
all_structs.append(HelmTrigger)
HelmTrigger.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'track', 'UTF8', None, ),  # 1
)
all_structs.append(DeploymentTrigger)
DeploymentTrigger.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'github', [GithubTrigger, None], None, ),  # 1
    (2, TType.STRUCT, 'image', [ImageTrigger, None], None, ),  # 2
    (3, TType.STRUCT, 'helm', [HelmTrigger, None], None, ),  # 3
)
all_structs.append(Chart)
Chart.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'reference', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'dir', 'UTF8', None, ),  # 2
    (3, TType.STRING, 'version', 'UTF8', None, ),  # 3
)
all_structs.append(Release)
Release.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'name', 'UTF8', None, ),  # 1
    (2, TType.STRUCT, 'chart', [Chart, None], None, ),  # 2
    (3, TType.STRING, 'valuesFile', 'UTF8', None, ),  # 3
    (4, TType.LIST, 'values', (TType.STRUCT, [ChartValue, None], False), None, ),  # 4
    (5, TType.STRUCT, 'trigger', [DeploymentTrigger, None], None, ),  # 5
    (6, TType.LIST, 'triggers', (TType.STRUCT, [DeploymentTrigger, None], False), None, ),  # 6
    (7, TType.BOOL, 'skipDefaultValues', None, None, ),  # 7
    (8, TType.LIST, 'resourceFiles', (TType.STRING, 'UTF8', False), None, ),  # 8
)
all_structs.append(KubernetesResourceRef)
KubernetesResourceRef.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'kind', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'name', 'UTF8', None, ),  # 2
)
all_structs.append(Releases)
Releases.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'resourceFiles', (TType.STRING, 'UTF8', False), None, ),  # 1
    (2, TType.LIST, 'releases', (TType.STRUCT, [Release, None], False), None, ),  # 2
    (3, TType.LIST, 'resourceDependencies', (TType.STRUCT, [KubernetesResourceRef, None], False), None, ),  # 3
)
all_structs.append(GkeProvider)
GkeProvider.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'project', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'clusterName', 'UTF8', None, ),  # 2
    (3, TType.STRING, 'zone', 'UTF8', None, ),  # 3
)
all_structs.append(MinikubeProvider)
MinikubeProvider.thrift_spec = (
)
all_structs.append(Provider)
Provider.thrift_spec = (
    None,  # 0
    (1, TType.STRUCT, 'gke', [GkeProvider, None], None, ),  # 1
    (2, TType.STRUCT, 'minikube', [MinikubeProvider, None], None, ),  # 2
)
all_structs.append(ClusterParameter)
ClusterParameter.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'name', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'value', 'UTF8', None, ),  # 2
)
all_structs.append(Cluster)
Cluster.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'name', 'UTF8', None, ),  # 1
    (2, TType.STRUCT, 'provider', [Provider, None], None, ),  # 2
    (3, TType.LIST, 'parameters', (TType.STRUCT, [ClusterParameter, None], False), None, ),  # 3
)
all_structs.append(Environment)
Environment.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'name', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'clusterName', 'UTF8', None, ),  # 2
    (3, TType.STRING, 'kubeNamespace', 'UTF8', None, ),  # 3
    (4, TType.LIST, 'releasesFiles', (TType.STRING, 'UTF8', False), None, ),  # 4
    (5, TType.STRING, 'defaultValuesFile', 'UTF8', None, ),  # 5
    (6, TType.LIST, 'defaultValues', (TType.STRUCT, [ChartValue, None], False), None, ),  # 6
)
all_structs.append(HelmRepo)
HelmRepo.thrift_spec = (
    None,  # 0
    (1, TType.STRING, 'name', 'UTF8', None, ),  # 1
    (2, TType.STRING, 'url', 'UTF8', None, ),  # 2
    (3, TType.STRING, 'caFile', 'UTF8', None, ),  # 3
    (4, TType.STRING, 'certFile', 'UTF8', None, ),  # 4
    (5, TType.STRING, 'keyFile', 'UTF8', None, ),  # 5
)
all_structs.append(KubecdConfig)
KubecdConfig.thrift_spec = (
    None,  # 0
    (1, TType.LIST, 'clusters', (TType.STRUCT, [Cluster, None], False), None, ),  # 1
    (2, TType.LIST, 'environments', (TType.STRUCT, [Environment, None], False), None, ),  # 2
    (3, TType.LIST, 'helmRepos', (TType.STRUCT, [HelmRepo, None], False), None, ),  # 3
)
fix_spec(all_structs)
del all_structs



from typing import List

from semantic_version import Version, validate, Spec


def is_semver(version: str) -> bool:
    try:
        parse(version)
        return True
    except ValueError:
        return False


def normalize(version: str) -> str:
    return version[1:] if version.startswith('v') else version


def parse(version: str) -> Version:
    version = normalize(version)
    return Version.coerce(version)


def filter_semver_tags(tags: List[str]) -> List[str]:
    """
    Filter a list of tags to only the tags that follow a semantic versioning scheme.
    :param tags:  input tags
    :return:  those input tags that are semver
    """
    output = []
    for tag in tags:
        if validate(normalize(tag)):
            output.append(tag)
    return output


def best_upgrade(current: Version, candidates: List[Version], track: str='MajorVersion'):
    if track == 'PatchLevel':
        spec = Spec('>{current},<{next_minor}'.format(current=str(current), next_minor=str(current.next_minor())))
    elif track == 'MinorVersion':
        spec = Spec('>{current},<{next_minor}'.format(current=str(current), next_minor=str(current.next_major())))
    elif track == 'MajorVersion':
        spec = Spec('>{current}'.format(current=str(current)))
    else:
        raise ValueError('unsupported "track": {track}'.format(track=track))
    return spec.select(candidates)


def is_wanted_upgrade(current: Version, candidate: Version, track: str='MajorVersion'):
    best = best_upgrade(current, [candidate], track)
    return best == candidate



import json
import os
from typing import Dict
from unittest.mock import patch

import pytest

from .. import updates as sut

TEST_DOCKER_HUB_TAGS_URL = 'https://registry.hub.docker.com/v2/repositories/confluentinc/cp-kafka/tags?page_size=256'
TEST_REGISTRY_TAGS_URL = 'https://docker.example.net:5001/v2/example/application/tags/list'
TEST_REGISTRY_MANIFESTS_URL_PREFIX = 'https://docker.example.net:5001/v2/example/application/manifests/'


def local_file(file_name: str):
    with open(os.path.join(os.path.dirname(__file__), 'testdata', file_name)) as f:
        return f.read()


def test_parse_image_repo():
    assert sut.parse_image_repo('centos') == ('docker.io', 'centos')
    assert sut.parse_image_repo('gcr.io/foo') == ('gcr.io', 'foo')


def mocked_requests_get(*args, **kwargs):
    class MockResponse:
        def __init__(self, status, response):
            self.status = status
            self.response = response

        def json(self):
            return json.loads(self.response)

    if args[0] == TEST_DOCKER_HUB_TAGS_URL:
        return MockResponse(200, local_file('docker-hub-kafka-tags.json'))
    elif args[0] == TEST_REGISTRY_TAGS_URL:
        return MockResponse(200, local_file('registry-v2-tags-list.json'))
    elif args[0].startswith(TEST_REGISTRY_MANIFESTS_URL_PREFIX):
        l = len(TEST_REGISTRY_MANIFESTS_URL_PREFIX)
        tag = args[0][l:]
        return MockResponse(200, local_file('registry-v2-manifests-{tag}.json'.format(tag=tag)))
    else:
        pytest.fail("unexpected requests.get url: {url}".format(url=args[0]))


def test_parse_docker_timestamp():
    ts = sut.parse_docker_timestamp('1970-01-01T02:00:00Z')
    assert ts == 7200, 'ts is {ts}'.format(ts=ts)


@patch('subprocess.check_output')
def test_get_tags_for_gcr_image(subprocess_mock, gcr_airflow_tags, gcr_expected_tags):
    subprocess_mock.return_value = gcr_airflow_tags.encode('utf-8')
    tags = sut.get_tags_for_gcr_image('us.gcr.io', 'gcp-project/airflow')
    assert tags == gcr_expected_tags


@patch('requests.get', side_effect=mocked_requests_get)
def test_get_tags_for_docker_hub_image(mock_get, docker_hub_expected_tags):
    tags = sut.get_tags_for_docker_hub_image('confluentinc/cp-kafka')
    assert tags == docker_hub_expected_tags


@patch('requests.get', side_effect=mocked_requests_get)
def test_get_tags_for_docker_v2_registry(mock_get, docker_registry_v2_expected_tags):
    tags = sut.get_tags_for_docker_v2_registry('docker.example.net:5001', 'example/application')
    print(tags)
    assert tags == docker_registry_v2_expected_tags


def test_get_newest_matching_tag(docker_hub_expected_tags):
    ver = '3.1.1'
    ts = docker_hub_expected_tags[ver]
    patch_update = sut.get_newest_matching_tag(ver, docker_hub_expected_tags, 'PatchLevel', ts)
    assert patch_update == '3.1.2'
    minor_update = sut.get_newest_matching_tag(ver, docker_hub_expected_tags, 'MinorVersion', ts)
    assert minor_update == '3.3.1'
    major_update = sut.get_newest_matching_tag(ver, docker_hub_expected_tags, 'MajorVersion', ts)
    assert major_update == '4.0.0'
    tags = {'1.0.0': 0, '1.0.0-0': 0}
    assert '1.0.0' == sut.get_newest_matching_tag('0.9', tags, 'MajorVersion', 0)
    tags = {'0.9': 2, '1.0': 1}
    assert '0.9' == sut.get_newest_matching_tag('0.9', tags, 'Newest', 0)


@pytest.fixture
def gcr_airflow_tags():
    return local_file('gcr-airflow-tags.json')


@pytest.fixture
def gcr_expected_tags():
    return {
        '1.8': 1512057463,
        '1.8.2': 1513690388
    }


@pytest.fixture
def registry_v2_tags_list():
    return local_file('registry-v2-tags-list.json')


@pytest.fixture
def registry_v2_manifests_test():
    return local_file('registry-v2-manifests-test.json')


@pytest.fixture
def docker_hub_kafka_tags():
    return local_file('docker-hub-kafka-tags.json')


@pytest.fixture
def docker_hub_expected_tags() -> Dict[str, int]:
    return {
        '3.0.0': 1469727557,
        '3.0.1': 1476822695,
        '3.0.1-1': 1473205014,
        '3.0.1-2': 1475191913,
        '3.0.1-3': 1476822692,
        '3.1.0': 1478710368,
        '3.1.0-1': 1478710365,
        '3.1.1': 1479253679,
        '3.1.1-1': 1479253675,
        '3.1.2': 1488481902,
        '3.1.2-1': 1485817659,
        '3.1.2-2': 1488481896,
        '3.2.0': 1490322876,
        '3.2.0-1': 1488590903,
        '3.2.0-2': 1488926205,
        '3.2.0-3': 1490307378,
        '3.2.0-4': 1490322873,
        '3.2.1': 1494629055,
        '3.2.1-4': 1493831069,
        '3.2.1-5': 1493922378,
        '3.2.1-6': 1494629053,
        '3.2.2': 1498523247,
        '3.2.2-1': 1498523244,
        '3.3.0': 1501532647,
        '3.3.0-1': 1501532644,
        '3.3.1': 1510600895,
        '4.0.0': 1515007897,
        '4.0.0-2': 1515007895,
        '4.0.0-3': 1518730912,
        'e618702': 1469648565,
        'latest': 1518730914
    }


@pytest.fixture
def docker_registry_v2_expected_tags() -> Dict[str, int]:
    return {
        'test':   1423821895,
        'latest': 1437468476,
    }



import json
import subprocess
from collections import defaultdict, OrderedDict
from typing import Tuple, Dict, List
import dateutil.parser
import logging

import requests
import semantic_version

from . import helm
from . import semver
from .model import Environment, Release
from .helm import key_is_in_values, lookup_value

logger = logging.getLogger(__name__)


def parse_docker_timestamp(timestamp: str) -> int:
    """
    Parse a timestamp of the format "2015-02-13T10:04:55.62062787Z" and
    return unix time
    :param timestamp:
    :return: unix timestamp (seconds since 1970-01-01 00:00:00 UTC)
    """
    return int(dateutil.parser.parse(timestamp).timestamp())


def parse_image_repo(repo: str) -> Tuple[str, str]:
    """
    If 'repo' contains a host prefix, return a tuple of that and the rest.
    If no host is found, we use "docker.io" instead.
    :param repo: full repository string
    :return: a tuple of (registry host, repository)
    """
    elem = repo.split('/')
    if len(elem) > 1 and '.' in elem[0]:
        return elem[0], '/'.join(elem[1:])
    return 'docker.io', repo


def get_tags_for_gcr_image(registry: str, repo: str) -> OrderedDict:
    full_repo = registry + '/' + repo
    gcp_project = repo.split('/')[0]
    cmd = ['gcloud', 'container', 'images', 'list-tags', full_repo, '--project', gcp_project, '--format', 'json']
    gcr_response = json.loads(subprocess.check_output(cmd).decode('utf-8'))
    response = OrderedDict()
    for img_tag in gcr_response:
        if 'timestamp' in img_tag:
            timestamp = parse_docker_timestamp(img_tag['timestamp']['datetime'])
        else:
            timestamp = 0
        if 'tags' in img_tag:
            # print(img_tag)
            for tag in img_tag['tags']:
                response[tag] = timestamp
    return response


def get_tags_for_docker_hub_image(repo: str) -> OrderedDict:
    response = OrderedDict()
    if '/' not in repo:
        repo = 'library/' + repo
    api_url = 'https://registry.hub.docker.com/v2/repositories/{repo}/tags?page_size=256'.format(repo=repo)
    docker_response = requests.get(api_url).json()
    if 'results' in docker_response:
        for item in docker_response['results']:
            response[item['name']] = parse_docker_timestamp(item['last_updated'])
    return response


def get_tags_for_docker_v2_registry(registry: str, repo: str) -> OrderedDict:
    response = OrderedDict()
    api_url = 'https://{registry}/v2/{repo}/tags/list'.format(registry=registry, repo=repo)
    registry_response = requests.get(api_url).json()
    if 'tags' in registry_response:
        for tag in registry_response['tags']:
            mf_url = 'https://{registry}/v2/{repo}/manifests/{tag}'.format(registry=registry, repo=repo, tag=tag)
            mf_response = requests.get(mf_url).json()
            if 'history' in mf_response and len(mf_response['history']) > 0:
                v1_compat = json.loads(mf_response['history'][0]['v1Compatibility'])
                response[tag] = parse_docker_timestamp(v1_compat['created'])
    return response


def get_tags_for_image(full_repo: str) -> OrderedDict:
    registry, repo = parse_image_repo(full_repo)
    if registry.endswith('gcr.io'):
        tags = get_tags_for_gcr_image(registry, repo)
    elif registry == 'docker.io':
        tags = get_tags_for_docker_hub_image(repo)
    else:
        tags = get_tags_for_docker_v2_registry(registry, repo)
    return tags


def filter_semver_tags(tags: List[str]) -> List[str]:
    """
    Filter a list of tags to only the tags that follow a semantic versioning scheme.
    :param tags:  input tags
    :return:  those input tags that are semver
    """
    output = []
    for tag in tags:
        if semantic_version.validate(semver.normalize(tag)):
            output.append(tag)
    return output


def get_newest_matching_tag(tag: str, tags: Dict[str, int], track: str, tag_timestamp: int = 0):
    if track == 'Newest':
        # filter out tags that are not newer than the current one, and ignore "latest"
        tags = {k: v for k, v in tags.items() if k != 'latest' and v > tag_timestamp}
        sorted_tags = sorted(tags, key=tags.__getitem__, reverse=True)  # sort by descending timestamp
        return sorted_tags[0] if len(sorted_tags) > 0 else None
    # If not "Newest", it is one of the semver variants
    candidates = filter_semver_tags(list(tags.keys()))  # filter out non-semver tags
    versions = {}
    for v in candidates:
        parsed = semver.parse(v)
        versions[parsed] = v
    current = semver.parse(tag)
    best = semver.best_upgrade(current, list(versions.keys()), track)
    if best is None:
        return None
    return versions[best]


class ImageUpdate(object):
    """
    An object representing an image with an available update. Just type safety convenience.
    """
    def __init__(self, new_tag: str, tag_value: str, release: Release, old_tag: str=None, image_repo: str=None):
        self.old_tag = old_tag
        self.new_tag = new_tag
        self.release = release
        self.tag_value = tag_value
        self.image_repo = image_repo


def release_wants_tag_update(release: Release, new_tag: str) -> List[ImageUpdate]:
    updates = []
    for trigger in release.triggers:
        if trigger.image is None or trigger.image.tagValue is None:
            continue
        values = helm.get_resolved_values(release, for_env=None, skip_value_from=True)
        tag_value = trigger.image.tagValue
        current_tag = lookup_value(tag_value, values)
        # if the current version is not semver, consider any value to be an update
        if not semver.is_semver(current_tag):
            updates.append(ImageUpdate(new_tag=new_tag, tag_value=tag_value, release=release))
            continue
        # if the new version is not semver, consider it an update only if track == Newest
        if not semver.is_semver(new_tag) and trigger.image.track == 'Newest':
            updates.append(ImageUpdate(new_tag=new_tag, tag_value=tag_value, release=release))
            continue
        if current_tag is not None and semver.is_wanted_upgrade(semver.parse(current_tag),
                                                                semver.parse(new_tag),
                                                                trigger.image.track):
            updates.append(ImageUpdate(new_tag=new_tag, tag_value=tag_value, release=release))
    return updates


def find_updates_for_release(release: Release, environment: Environment) -> Dict[str, List[ImageUpdate]]:
    updates = defaultdict(list)
    if release.triggers is None:
        return updates
    for trigger in release.triggers:
        if trigger.image is None:
            continue
        tag_value = trigger.image.tagValue
        repo_value = trigger.image.repoValue
        prefix_value = trigger.image.repoPrefixValue
        track = trigger.image.track
        values = helm.get_resolved_values(release, for_env=environment, skip_value_from=True)
        logger.debug('found trigger for image "%s" from value "%s": %s final values: %s',
                     lookup_value(repo_value, values),
                     tag_value,
                     json.dumps(trigger.image.__dict__),
                     json.dumps(values))
        if key_is_in_values(repo_value, values):
            image_repo = lookup_value(repo_value, values)
            if key_is_in_values(prefix_value, values):
                image_repo = lookup_value(prefix_value, values) + image_repo
            image_tag = lookup_value(tag_value, values)
            all_tags = get_tags_for_image(image_repo)
            tag_timestamp = all_tags[image_tag] if image_tag in all_tags else 0
            updated_tag = get_newest_matching_tag(image_tag, all_tags, track, tag_timestamp)
            if updated_tag is not None:
                logger.debug('found update for "%s": "%s" -> "%s"', image_repo, image_tag, updated_tag)
                updates[release.from_file].append(ImageUpdate(old_tag=image_tag,
                                                              new_tag=updated_tag,
                                                              release=release,
                                                              tag_value=tag_value,
                                                              image_repo=image_repo))
    logger.debug('find_updates_for_release: returning: %r', updates)
    return updates


def find_updates_for_releases(releases: List[Release], environment: Environment) -> Dict[str, List[ImageUpdate]]:
    env_updates = defaultdict(list)
    for release in releases:
        logger.info('checking updates for release: {env}/{release}'.format(env=environment.name, release=release.name))
        image_updates = find_updates_for_release(release, environment)
        for file, updates in image_updates.items():
            env_updates[file].extend(updates)
    logger.debug('find_updates_for_env: returning: %r', env_updates)
    return env_updates


def find_updates_for_env(environment: Environment) -> Dict[str, List[ImageUpdate]]:
    env_updates = defaultdict(list)
    for release in environment.all_releases:
        logger.info('checking updates for release: {env}/{release}'.format(env=environment.name, release=release.name))
        image_updates = find_updates_for_release(release, environment)
        for file, updates in image_updates.items():
            env_updates[file].extend(updates)
    logger.debug('find_updates_for_env: returning: %r', env_updates)
    return env_updates



import re
import subprocess
from os import path
from typing import List, Union

import ruamel.yaml

from . import model
from .gen_py import ttypes
from .utils import resolve_file_path

import logging

logger = logging.getLogger(__name__)


def inspect(chart_reference: str, chart_version: str) -> str:
    cmd = ['helm', 'inspect', chart_reference, '--version', chart_version]
    output = subprocess.check_output(cmd).decode('utf-8').split("\n---\n")[1]
    return output


def repo_setup_commands(repos: List[ttypes.HelmRepo]) -> List[List[str]]:
    cmds = [['helm', 'repo', 'add', repo.name, repo.url] for repo in repos]
    cmds.append(['helm', 'repo', 'update'])
    return cmds


def kubectl_apply_command(resource_files: List[str], dry_run: bool, env_name: str) -> List[str]:
    cmd = ['kubectl', '--context', 'env:{}'.format(env_name), 'apply']
    if dry_run:
        cmd.append('--dry-run')
    for resource_file in resource_files:
        cmd.extend(['-f', resource_file])
    return cmd


def deploy_commands(env: model.Environment, dry_run=False, debug=False, limit_to_releases=None) -> List[List[str]]:
    commands = []
    if limit_to_releases is None:
        commands.append(kubectl_apply_command(env.all_resource_files, dry_run, env.name))
    else:
        for r in limit_to_releases:
            if env.named_release(r) is None:
                raise ValueError('no such release in env:{}: {}'.format(env.name, r))
    for release in env.all_releases:
        if limit_to_releases is None or release.name in limit_to_releases:
            rel_file = release.from_file
            if release.chart:
                commands.append(generate_helm_install_argv(
                    release, env, release_file=rel_file, dry_run=dry_run, debug=debug))
            elif release.resourceFiles:
                abs_files = [resolve_file_path(p, rel_file) for p in release.resourceFiles]
                commands.append(kubectl_apply_command(abs_files, dry_run, env.name))
    return commands


def generate_helm_base_argv(env: model.Environment) -> List[str]:
    return ['helm', '--kube-context', 'env:{}'.format(env.name)]


def generate_helm_values_argv(rel: model.Release, env: model.Environment, release_file: str) -> List[str]:
    argv = []
    if not rel.skipDefaultValues:
        if env.defaultValuesFile:
            def_val_file = resolve_file_path(env.defaultValuesFile, release_file)
            argv.extend(['--values', def_val_file])
        if env.defaultValues:
            argv.append('--set')
            argv.append(','.join(['='.join(resolve_value(x, env)) for x in env.defaultValues]))
    if rel.valuesFile:
        val_file = resolve_file_path(rel.valuesFile, release_file)
        argv.extend(['--values', val_file])
    if rel.values:
        argv.append('--set')
        argv.append(','.join(['='.join(resolve_value(x, env)) for x in rel.values]))
    return argv


def generate_helm_chart_args(rel: model.Release, release_file: str) -> List[str]:
    if rel.chart.reference is None:
        dir = resolve_file_path(rel.chart.dir, release_file)
        if not path.exists(dir):
            raise ValueError('{}: release "{}" chart.dir "{}" does not exist'.format(
                release_file, rel.name, dir))
        return [dir]
    return [rel.chart.reference, '--version', rel.chart.version]


def generate_helm_diff_argv(rel: model.Release, env: model.Environment, release_file: str) -> List[str]:
    argv = generate_helm_base_argv(env)
    argv.extend(['diff', 'upgrade', rel.name, generate_helm_chart_args(rel, release_file)])
    argv.extend(generate_helm_values_argv(rel, env, release_file))
    return argv


def generate_helm_install_argv(rel: model.Release,
                               env: model.Environment,
                               release_file: str,
                               dry_run: bool = False,
                               debug: bool = False) -> List[str]:
    chart_args = generate_helm_chart_args(rel, release_file)
    argv = generate_helm_base_argv(env)
    argv.extend(['upgrade', rel.name])
    argv.extend(chart_args)
    argv.extend(['-i', '--namespace', env.kubeNamespace])
    argv.extend(generate_helm_values_argv(rel, env, release_file))
    if dry_run:
        argv.append('--dry-run')
    if debug:
        argv.append('--debug')
    return argv


def resolve_gce_address_value(address: ttypes.GceAddressValueRef, env: model.Environment):
    cmd = ['gcloud', 'compute', 'addresses', 'describe', address.name, '--format', 'value(address)']
    provider = env.cluster.provider
    cmd.extend(['--project', provider.gke.project])
    if address.isGlobal:
        cmd.append('--global')
    else:
        cmd.extend(['--region', re.sub(r'-[a-z]$', '', provider.gke.zone)])
    return subprocess.check_output(cmd).decode('utf-8').strip()


def resolve_value(value: ttypes.ChartValue, env: model.Environment, skip_value_from=False) -> tuple:
    v = value.value
    if value.valueFrom and not skip_value_from:
        if value.valueFrom.gceResource:
            if value.valueFrom.gceResource.address:
                v = resolve_gce_address_value(value.valueFrom.gceResource.address, env)
    return value.key, v


def merge_values(from_dict: dict, onto_dict: dict) -> dict:
    for key, value in from_dict.items():
        if isinstance(value, dict):
            # get node or create one
            node = onto_dict.setdefault(key, {})
            if isinstance(node, dict):
                merge_values(value, node)
        else:
            onto_dict[key] = value
    return onto_dict


def load_values_file(file_name: str) -> dict:
    with open(file_name) as f:
        return ruamel.yaml.safe_load(f)


def values_list_to_dict(values: List[ttypes.ChartValue], env: model.Environment, skip_value_from=False) -> dict:
    """
    Convert a list of ChartValue objects with keys on the form ``"foo.bar"``
    to a nested dictionary like ``{"foo": {"bar": ...}}``.
    :param values: list of ChartValue objects
    :param env: current environment
    :param skip_value_from: whether to skip resolving "valueFrom" entries (which will invoke
                            commands or API calls)
    :return:
    """
    def val_to_dict(key: List[str], value) -> dict:
        if len(key) == 1:
            return {key[0]: value}
        return {key[0]: val_to_dict(key[1:], value)}

    output = {}
    for value_obj in values:
        k, v = resolve_value(value_obj, env, skip_value_from=skip_value_from)
        output = merge_values(from_dict=val_to_dict(k.split('.'), v), onto_dict=output)
    return output


def lookup_value(key: Union[List[str], str], values: dict):
    if type(key) == str:
        key = key.split('.')
    if len(key) > 0 and len(values) > 0:
        if key[0] in values:
            if len(key) == 1:
                return values[key[0]]
            return lookup_value(key[1:], values[key[0]])
    return None


def key_is_in_values(key: Union[List[str], str], values: dict) -> bool:
    if type(key) == str:
        key = key.split('.')
    if len(key) > 0 and len(values) > 0:
        if len(key) == 1:
            return key[0] in values
        if key[0] in values:
            return key_is_in_values(key[1:], values[key[0]])
    return False


def get_resolved_values(release: model.Release, for_env: Union[model.Environment, None], skip_value_from=False) -> dict:
    # 1. get values from chart if in a local dir
    # 2. merge with valuesFile:
    # 3. merge with values:
    values = {}
    if release.chart.dir:
        values_file = resolve_file_path(path.join(release.chart.dir, 'values.yaml'), relative_to_file=release._from_file)
        if path.exists(values_file):
            values = merge_values(from_dict=load_values_file(values_file), onto_dict=values)
    elif release.chart.reference:
        # "helm inspect" outputs a two-document yaml block, where the second is the parsed default values.yaml
        output = inspect(release.chart.reference, release.chart.version)
        chart_default_values = ruamel.yaml.safe_load(output)
        values = merge_values(from_dict=chart_default_values, onto_dict=values)
    if for_env is not None and len(for_env.defaultValues) > 0:
        default_values = values_list_to_dict(for_env.defaultValues, for_env, skip_value_from=skip_value_from)
        values = merge_values(from_dict=default_values, onto_dict=values)
    if release.valuesFile:
        values_file = resolve_file_path(release.valuesFile, relative_to_file=release._from_file)
        values = merge_values(from_dict=load_values_file(values_file), onto_dict=values)
    if for_env is not None and release.values:
        values = merge_values(from_dict=values_list_to_dict(release.values, for_env, skip_value_from=skip_value_from),
                              onto_dict=values)
    return values



import argparse
import os
import subprocess
import sys
from typing import List

import argcomplete
import logging
import ruamel.yaml
import ruamel.yaml.error
from argcomplete import FilesCompleter
from blessings import Terminal
from ruamel.yaml import YAMLError

from kubecd import __version__, helm, updates, github
from kubecd import model
from kubecd.thriftutils import SchemaError
from kubecd.updates import find_updates_for_env, find_updates_for_releases

t = Terminal()
logger = logging.getLogger(__name__)


class CliError(Exception):
    pass


def parser(prog='kcd') -> argparse.ArgumentParser:
    p = argparse.ArgumentParser(prog=prog)
    p.add_argument(
        '-f', '--config-file',
        help='KubeCD config file file (default $KUBECD_CONFIG or "environments.yaml")',
        metavar='FILE',
        default=os.getenv('KUBECD_CONFIG', 'environments.yaml')).completer = FilesCompleter(
            allowednames=('.yaml', '.yml'), directories=False)
    p.add_argument(
        '--version',
        help='Show version and exit.',
        action='version',
        version='kubecd ' + __version__)
    p.add_argument('--verbose', '-v', help='Increase verbosity level', action='count', default=0)

    s = p.add_subparsers(dest='command', title='Subcommands', description='Use one of these sub-commands:')

    apply = s.add_parser('apply', help='apply changes to Kubernetes')
    apply.add_argument('--dry-run', '-n', help='dry run mode, only print commands', action='store_true', default=False)
    apply.add_argument('--debug', help='run helm with --debug', action='store_true', default=False)
    apply.add_argument('--releases', '-r', help='apply only these releases', action='append')
    apply.add_argument('--all-environments', '-a', help='apply all environments', action='store_true', default=False)
    apply.add_argument('env', nargs='?',
                       help='name of environment to apply, must be specified unless --all-environments is')
    apply.set_defaults(func=apply_env)

    # diff = s.add_parser('diff', help='show diffs between running and git release')
    # diff.add_argument('--releases', '-r', help='which releases to diff', action='append')
    # diff.add_argument('env', nargs='?', help='name of environment')
    # diff.set_defaults(func=diff_release)

    poll_p = s.add_parser('poll', help='poll for new images in registries')
    poll_p.add_argument('--patch', '-p', help='patch releases.yaml files with updated version', action='store_true')
    poll_p.add_argument('--releases', '-r', help='poll this specific release', action='append')
    poll_p.add_argument('--image', '-i', help='poll releases using this image')
    poll_p.add_argument('env', nargs='?', help='name of environment to poll')
    poll_p.set_defaults(func=poll_registries)

    dump_p = s.add_parser('dump', help='dump commands for one or all environments')
    dump_p.add_argument('--all-environments', '-a', help='dump all environments', action='store_true', default=False)
    dump_p.add_argument('env', nargs='?', help='name of environment to dump')
    dump_p.set_defaults(func=dump_env)

    list_p = s.add_parser('list', help='list environments or releases')
    list_p.add_argument('kind', help='what to list', choices=['env', 'release'])
    list_p.set_defaults(func=list_kind)

    indent_p = s.add_parser('indent', help='canonically indent YAML files')
    indent_p.add_argument('files', nargs='+', help='file[s] to indent')
    indent_p.set_defaults(func=indent_file)

    observe = s.add_parser('observe', help='observe a new image version')
    observe.add_argument('--image', '-i', help='the image, including tag', metavar='IMAGE:TAG')
    observe.add_argument('--patch', help='patch release files with updated tags',
                         action='store_true', default=False)
    observe.add_argument('--submit-pr', help='submit a pull request with the updated tags',
                         action='store_true', default=False)
    observe.set_defaults(func=observe_new_image)

    completion_p = s.add_parser('completion', help='print shell completion script')
    completion_p.set_defaults(func=print_completion, prog=prog)
    return p


def diff_release(config_file: str, releases: List[str], env: str, **kwargs):
    kcd_config = load_model(config_file)
    e = kcd_config.env_by_name(env)
    for release in releases:
        r = e.named_release(release)
        cmd = helm.generate_helm_diff_argv(r, e, r.from_file)
        s = subprocess.call(cmd)
        if s != 0:
            raise CliError('Command "{cmd}" exited with status {status}'.format(cmd=' '.join(cmd), status=s))


def observe_new_image(config_file: str, image: str, patch: bool, submit_pr: bool, **kwargs):
    kcd_config = load_model(config_file)
    image_repo, image_tag = image.split(':')
    image_index = kcd_config.image_index
    patched_files = []
    if image_repo in image_index:
        for release in image_index[image_repo]:
            update_list = updates.release_wants_tag_update(release, image_tag)
            for update in update_list:
                print('release: {release} tagValue: {tag}'.format(release=update.release.name, tag=update.tag_value))
            if len(update_list) > 0 and patch:
                patch_releases_file(release.from_file, update_list)
                patched_files.append(release.from_file)
    if submit_pr and len(patched_files) > 0:
        github.pr_from_files(patched_files, 'Observed image {image}'.format(image=image))


def print_completion(prog, **kwargs):
    shell = os.path.basename(os.getenv('SHELL'))
    if shell == 'bash' or shell == 'tcsh':
        sys.stdout.write(argcomplete.shellcode(prog, shell=shell))


def apply_env(config_file, dry_run, debug, all_environments=False, env=None, releases=None, **kwargs):
    target_envs = one_or_all_envs(env, all_environments, file_name=config_file)
    commands_to_run = model.config().init_commands()
    for environment in target_envs:
        logger.info('Collecting commands for environment "%s"', environment.name)
        init_cmds = environment.init_commands(dry_run=dry_run)
        try:
            deploy_cmds = helm.deploy_commands(environment, dry_run=dry_run, limit_to_releases=releases, debug=debug)
        except ValueError as e:
            raise CliError(str(e))
        if len(deploy_cmds) > 0:
            commands_to_run.extend(init_cmds)
            commands_to_run.extend(deploy_cmds)
    for cmd in commands_to_run:
        print('{t.yellow}{cmd}{t.normal}'.format(cmd=' '.join(cmd), t=t))
        logger.debug('Executing: "%s"', ' '.join(cmd))
        cmd_status = subprocess.call(cmd)
        if cmd_status != 0:
            raise CliError('Command "{cmd}" exited with non-0 status {status}'.format(cmd=cmd, status=cmd_status))


def dump_env(config_file, all_environments=False, env=None, **kwargs):
    target_envs = one_or_all_envs(env, all_environments, file_name=config_file)
    for cmd in model.config().init_commands():
        print('{t.yellow}{cmd}{t.normal}'.format(cmd=' '.join(cmd), t=t))
    for environment in target_envs:
        print('{t.green}Environment:{t.normal} {env_name}'.format(env_name=environment.name, t=t))
        for cmd in environment.init_commands(dry_run=False):
            print('{t.yellow}{cmd}{t.normal}'.format(cmd=' '.join(cmd), t=t))
        for cmd in helm.deploy_commands(environment, dry_run=False):
            print(' '.join(cmd))
        print('')


def list_kind(config_file, kind, **kwargs):
    if kind == 'env' or kind == 'environment':
        for environment in load_model(config_file):
            print(environment.name)
    elif kind == 'release':
        for environment in load_model(config_file):
            for release in environment.all_releases:
                print('{env}/{release}'.format(env=environment.name, release=release.name))
    else:
        raise CliError('unknown kind "{}"'.format(kind))


def poll_registries(config_file, all_environments=False, env=None, releases=None, patch=False, **kwargs):
    target_envs = one_or_all_envs(env, all_environments, file_name=config_file)
    for environment in target_envs:
        if releases is None:
            logger.info('polling environment: "%s"', environment.name)
            file_updates = find_updates_for_env(environment)
        else:
            release_objs = [environment.named_release(r) for r in releases]
            logger.info('polling env:%s releases: %s', environment.name, ' '.join(releases))
            file_updates = find_updates_for_releases(release_objs, environment)
        for release_file, update_list in file_updates.items():
            for update in update_list:
                print(
                    '{env}/{release}:\n\tfile: {file}\n\timage: {image}\n\ttag: {tag_from} -> {tag_to}'.format(
                        env=environment.name,
                        file=release_file,
                        release=update.release.name,
                        image=update.image_repo,
                        tag_from=update.old_tag,
                        tag_to=update.new_tag,
                    ))
            if patch:
                patch_releases_file(release_file, update_list)


def patch_releases_file(releases_file: str, updates_list: List[updates.ImageUpdate]):
    logger.debug('loading releases file: "%s"', releases_file)
    mod_yaml = load_yaml(releases_file)
    for update in updates_list:
        for mod_rel in mod_yaml['releases']:
            if mod_rel['name'] == update.release.name:
                if 'values' not in mod_rel:
                    mod_rel['values'] = []
                found_val = False
                for yv in mod_rel['values']:
                    if yv['key'] == update.tag_value:
                        logger.debug('patching value "{}"'.format(update.tag_value))
                        yv['value'] = update.new_tag
                        found_val = True
                        break
                if not found_val:
                    mod_rel['values'].append({'key': update.tag_value, 'value': update.new_tag})
    logger.debug('saving patched file: {file}'.format(file=releases_file))
    save_yaml(mod_yaml, releases_file)


def indent_file(files, **kwargs):
    for filename in files:
        try:
            save_yaml(load_yaml(filename), filename)
        except YAMLError as e:
            raise CliError('invalid YAML file "{}": {}'.format(filename, str(e)))


def load_envs(file_name: str) -> List[model.Environment]:
    try:
        load_model(file_name)
        return model.as_list()
    except ruamel.yaml.error.YAMLError as e:
        raise CliError('could not read "{}": {}'.format(file_name, str(e)))


def one_or_all_envs(env_name: str, all_envs: bool, file_name: str) -> List[model.Environment]:
    load_envs(file_name)
    if env_name is None and all_envs is False:
        raise CliError('need to specify either an environment name or --all')
    if env_name is not None:
        try:
            return [model.get_environment(env_name)]
        except KeyError:
            raise CliError('no such environment: {}'.format(env_name))
    return model.as_list()


def load_model(file_name: str):
    try:
        return model.load(file_name)
    except FileNotFoundError as e:
        raise CliError('file not found: {}'.format(e.filename))
    except SchemaError as e:
        raise CliError(e)


def load_yaml(file_name: str):
    yaml = ruamel.yaml.YAML(typ='rt')
    yaml.indent(sequence=4, offset=2)
    yaml.preserve_quotes = True
    with open(file_name, 'r') as rf:
        return yaml.load(rf)


def save_yaml(mod_yaml, file_name: str):
    yaml = ruamel.yaml.YAML(typ='rt')
    yaml.indent(sequence=4, offset=2)
    yaml.preserve_quotes = True
    tmp_file = file_name + '.new'
    with open(tmp_file, 'w') as wf:
        yaml.dump(mod_yaml, wf)
    os.rename(tmp_file, file_name)


def verbose_log_level(v):
    if v == 0:
        return logging.WARNING
    if v == 1:
        return logging.INFO
    return logging.DEBUG


def main():
    p = parser()
    argcomplete.autocomplete(p)
    args = p.parse_args()
    kwargs = args.__dict__
    if 'func' not in kwargs:
        p.print_help(sys.stderr)
        sys.exit(1)
    func = kwargs['func']
    del (kwargs['func'])
    logging.basicConfig(stream=sys.stderr,
                        format='{levelname} {message}',
                        style='{',
                        level=verbose_log_level(args.verbose))
    try:
        func(**kwargs)
    except CliError as e:
        print('{t.red}ERROR{t.normal}: {msg}'.format(msg=str(e), t=t), file=sys.stderr)


if __name__ == '__main__':
    main()



from collections import defaultdict
from os import path
from typing import List, Union, Dict

import logging

from . import helm
from .gen_py import ttypes
from .providers import generate_environment_init_command
from .thriftutils import load_yaml_with_schema
from .utils import resolve_file_path

_config = None


class ConfigError(BaseException):
    pass


class Release(ttypes.Release):
    # _from_file: str

    def __init__(self, data: ttypes.Release, from_file: str):
        self._from_file = from_file
        super(Release, self).__init__(**data.__dict__)
        issues = self._sanity_check()
        if len(issues) > 0:
            raise ValueError('release {} issues found: {}'.format(self.name, ', '.join(issues)))
        self._compat_update()

    def _compat_update(self):
        if self.trigger:
            self.triggers = [self.trigger]
            self.trigger = None

    def _sanity_check(self):
        issues = []
        if self.chart is None and self.resourceFiles is None:
            issues.append('must define either "chart" or "resourceFiles"')
        if self.chart is not None and self.resourceFiles is not None:
            issues.append('must define only one of "chart" or "resourceFiles"')
        if self.chart is not None:
            if self.chart.reference is not None and self.chart.version is None:
                issues.append('must have a chart.version'.format(self.chart.name))
        if self.trigger and self.triggers:
            issues.append('must define only one of "trigger" or "triggers"')
        return issues

    @property
    def from_file(self):
        return self._from_file


class Environment(ttypes.Environment):
    # _from_file: str
    # _all_releases: List[Release]
    # _all_resource_files: List[str]

    def __init__(self, data: ttypes.Environment, from_file: str):
        self._from_file = path.abspath(from_file)
        self._all_releases = []
        self._all_resource_files = []

        super(Environment, self).__init__(**data.__dict__)
        issues = self._sanity_check()
        if len(issues) > 0:
            raise ValueError('Issues found:\n\t{}'.format('\n\t'.join(issues)))
        for rel_file in self.releasesFiles:
            abs_file = resolve_file_path(rel_file, relative_to_file=self._from_file)
            releases = load_yaml_with_schema(abs_file, ttypes.Releases)
            if releases.resourceFiles is not None:
                logging.warning('%s: resourceFiles at the top level are deprecated, move them to the a release',
                                abs_file)
                abs_files = [resolve_file_path(x, relative_to_file=abs_file) for x in releases.resourceFiles]
                self._all_resource_files.extend(abs_files)
            for rel in releases.releases:
                self._all_releases.append(Release(rel, from_file=abs_file))

    def _sanity_check(self):
        counts = defaultdict(int)
        issues = []
        for env in self.all_releases:
            if env.name in counts:
                issues.append('duplicate environment name: {}'.format(env.name))
            counts[env.name] += 1
        return issues

    @property
    def all_releases(self) -> List[Release]:
        return self._all_releases

    def named_release(self, name: str) -> Union[Release, None]:
        for r in self._all_releases:
            if r.name == name:
                return r
        return None

    @property
    def all_resource_files(self):
        return self._all_resource_files

    def init_commands(self, dry_run=False):
        cluster = get_cluster(self.clusterName)
        return generate_environment_init_command(cluster, self, dry_run)

    @property
    def cluster(self):
        return get_cluster(self.clusterName)


class Cluster(ttypes.Cluster):
    def __init__(self, data: ttypes.Cluster, from_file: str):
        self._from_file = from_file
        super(Cluster, self).__init__(**data.__dict__)

    @property
    def from_file(self):
        return self._from_file


class KubecdConfig(ttypes.KubecdConfig):
    # _environments: List[Environment]
    # _index: Dict[str, Environment] = {}
    # _from_file: str

    def __init__(self, data: ttypes.KubecdConfig, from_file: str):
        self._environments = [Environment(x, from_file) for x in data.environments]
        self._clusters = [Cluster(x, from_file) for x in data.clusters]
        self._from_file = from_file
        self._env_index = {}
        self._cluster_index = {}
        self._image_index = None
        super(KubecdConfig, self).__init__(**data.__dict__)
        issues = self._sanity_check()
        if len(issues) > 0:
            raise ValueError('Issues found:\n\t{}'.format('\n\t'.join(issues)))
        for env in self._environments:
            self._env_index[env.name] = env
        for cluster in self._clusters:
            self._cluster_index[cluster.name] = cluster

    def __iter__(self):
        return iter(self._environments)

    def _sanity_check(self):
        counts = defaultdict(int)
        issues = []
        for env in self._environments:
            if env.name in counts:
                issues.append('duplicate environment name: {}'.format(env.name))
            counts[env.name] += 1
        return issues

    def env_by_name(self, name: str) -> Environment:
        return self._env_index[name]

    def cluster_by_name(self, name: str) -> Cluster:
        return self._cluster_index[name]

    def init_commands(self) -> List[List[str]]:
        commands = []
        if self.helmRepos:
            commands.extend(helm.repo_setup_commands(self.helmRepos))
        return commands

    @property
    def image_index(self) -> Dict[str, List[Release]]:
        index = defaultdict(list)
        if self._image_index is None:
            for env in self._environments:
                for rel in env.all_releases:
                    values = helm.get_resolved_values(rel, env, skip_value_from=True)
                    if rel.triggers:
                        for t in rel.triggers:
                            if t.image:
                                repo = helm.lookup_value(t.image.repoValue, values)
                                if repo is None:
                                    break
                                prefix = helm.lookup_value(t.image.repoPrefixValue, values)
                                if prefix is not None:
                                    repo = prefix + repo
                                index[repo].append(rel)
            self._image_index = index
        return self._image_index

    @property
    def from_file(self):
        return self._from_file


def load(file_name: str) -> KubecdConfig:
    global _config
    _config = KubecdConfig(load_yaml_with_schema(file_name, ttypes.KubecdConfig), from_file=file_name)
    return _config


def as_list() -> List[Environment]:
    if _config is None:
        raise ConfigError('environments not yet loaded')
    return [x for x in _config]


def get_cluster(cluster_name: str) -> Cluster:
    if _config is None:
        raise ConfigError('environments not yet loaded')
    return _config.cluster_by_name(cluster_name)


def get_environment(env_name: str) -> Environment:
    if _config is None:
        raise ConfigError('environments not yet loaded')
    return _config.env_by_name(env_name)


# def generate_helm_command(rel: Release, env: Environment, release_file: str, dry_run: bool = False) -> str:
#     return ' '.join([shlex.quote(arg) for arg in generate_helm_command_argv(rel, env, release_file, dry_run)])

def releases_subscribing_to_image(image: str):
    pass


def config() -> KubecdConfig:
    return _config




import os

from setuptools import setup, find_packages, Command

module_path = os.path.join(os.path.dirname(__file__), 'kubecd', '__init__.py')
version_line = [line for line in open(module_path)
                if line.startswith('__version__')][0]

__version__ = version_line.split('__version__ = ')[-1][1:][:-2]


class ReleaseCommand(Command):
    """ Run my command.
    """
    description = 'tag a new release'
    version = None
    user_options = [
        ('version=', 'v', 'new version'),
        ('force', 'f', 'force push version tag')
    ]

    def initialize_options(self):
        self.version = None
        self.force = False

    def finalize_options(self):
        if self.version is None:
            self.version = __version__

    def run(self):
        import subprocess
        tag = 'v' + self.version
        push_cmd = ["git", "push", "--tags"]
        if self.force:
            push_cmd.append("-f")
            print("Untagging %s ..." % tag)
            if subprocess.call(["git", "tag", "-d", tag]) != 0:
                raise Exception("'git untag' command failed")
        print("Tagging %s ..." % tag)
        if subprocess.call(["git", "tag", tag]) != 0:
            raise Exception("'git tag' command failed")
        print("Pushing tags ...")
        if subprocess.call(push_cmd) != 0:
            raise Exception("'git push --tags' command failed")


setup(
    name='kubecd',
    version=__version__,
    cmdclass={
        'release': ReleaseCommand
    },
    description='Kubernetes Continuous Deployment and Inventory Tool',
    url='http://github.com/zedge/kubecd',
    author='Stig Bakken',
    author_email='stig@zedge.net',
    license='Apache 2',
    packages=find_packages(),
    python_requires='>=3.5.0',
    install_requires=[
        'thrift==0.11.0',
        'ruamel.yaml~=0.15.37',
        'semantic_version~=2.6.0',
        'requests~=2.18.4',
        'argcomplete~=1.9.4',
        'argparse~=1.4.0',
        'blessings',
        'python-dateutil'
    ],
    tests_require=['pytest', 'pytest-cov'],
    entry_points={
        'console_scripts': [
            'kcd=kubecd.cli:main'
        ]
    }
)



from thrift.Thrift import TType


class SchemaError(BaseException):
    pass


def load_yaml_with_schema(yaml_file: str, schema):
    with open(yaml_file, 'r') as fd:
        from ruamel import yaml
        try:
            return to_thrift_object(yaml.safe_load(fd), schema, '')
        except SchemaError as e:
            raise SchemaError('{}: {}'.format(yaml_file, str(e))) from e


def to_thrift_object(in_dict: dict, schema, obj_path: str):
    if not isinstance(in_dict, dict):
        raise SchemaError('unexpected type at "{}", expected dict, found {}'.format(obj_path, type(in_dict).__name__))

    ctor_args = {}
    for field_spec in schema.thrift_spec:
        if field_spec is None:
            continue
        t_type = field_spec[1]
        t_field = field_spec[2]
        if t_field in in_dict:
            new_path = obj_path
            if len(new_path) > 0:
                new_path += '.'
            new_path += t_field
            ctor_args[t_field] = to_thrift_type(in_dict[t_field], t_type, field_spec[3], new_path)
    extra_keys = set(in_dict.keys()) - set(ctor_args.keys())
    if len(extra_keys) > 0:
        raise SchemaError('extraneous keys for "{}": {}'.format(schema.__name__, ', '.join(extra_keys)))

    return schema(**ctor_args)


def to_thrift_type(value, t_type, t_subtype, obj_path: str):
    """
    Convert a Python value to a Thrift value
    :param value: python value
    :param t_type:
    :param t_subtype:
    :param obj_path:
    :return: Thrift-ified version of the input value
    :raises SchemaError: if the input value does not match the schema
    """
    if t_type == TType.I08 or t_type == TType.I16 or t_type == TType.I32 or t_type == TType.I64:
        return int(value)
    elif t_type == TType.BOOL:
        return bool(value)
    elif t_type == TType.DOUBLE:
        return float(value)
    elif t_type == TType.STRING:
        # t_subtype is encoding
        return str(value)
    elif t_type == TType.STRUCT:
        return to_thrift_object(value, t_subtype[0], obj_path)
    elif t_type == TType.LIST:
        if not isinstance(value, list):
            raise SchemaError('value is not a list at %s' % obj_path)
        return to_thrift_list(value, t_subtype, obj_path)
    elif t_type == TType.MAP:
        if not isinstance(value, dict):
            raise SchemaError('value is not a dict at %s' % obj_path)
        return to_thrift_map(value, t_subtype, obj_path)
    else:
        raise SchemaError('unknown or unsupported thrift type %d at %s' % (t_type, obj_path))


def to_thrift_list(in_list: list, spec: tuple, list_path: str):
    return [to_thrift_type(v, spec[0], spec[1], '{0}[{1}]'.format(list_path, i)) for i, v in enumerate(in_list)]


def to_thrift_map(in_map: dict, spec: tuple, map_path: str):
    # spec: (TType.STRING, 'UTF8', TType.STRING, 'UTF8', False)
    return {
        to_thrift_type(k, spec[0], spec[1], map_path):
            to_thrift_type(v, spec[2], spec[3], '{0}[{1}]'.format(map_path, k)) for k, v in in_map.items()
    }



import subprocess
import time
from typing import List


def pr_from_files(files: List[str], message: str):
    branch_name = 'br%d' % time.time()
    subprocess.run(['git', 'checkout', '-b', branch_name], check=True)
    git_add = ['git', 'add']
    git_add.extend(files)
    # print('git_add_cmd:', git_add)
    subprocess.run(git_add, check=True)
    subprocess.run(['git', 'commit', '-m', message], check=True)
    subprocess.run(['git', 'push', 'origin', branch_name], check=True)
    subprocess.run(['hub', 'pull-request', '-m', message], check=True)
